@@ -50,7 +50,7 @@ static PyObject *checkPassword(PyObject *self, PyObject *args) {
     const char *pswd = NULL;
     const char *service = NULL;
     const char *default_realm = NULL;
-    const int verify = 1;
+    int verify = 1;
     int result = 0;
 
     if (!PyArg_ParseTuple(args, "ssss|b", &user, &pswd, &service, &default_realm, &verify)) {
@@ -163,13 +163,82 @@ static PyObject *authGSSClientClean(PyObject *self, PyObject *args) {
     return Py_BuildValue("i", AUTH_GSS_COMPLETE);
 }
 
-static PyObject *authGSSClientStep(PyObject *self, PyObject *args) {
+#if PY_MAJOR_VERSION >= 3
+void destruct_channel_bindings(PyObject* o) {
+    struct gss_channel_bindings_struct *channel_bindings = PyCapsule_GetPointer(o, NULL);
+#else
+void destruct_channel_bindings(void* o) {
+    struct gss_channel_bindings_struct *channel_bindings = (struct gss_channel_bindings_struct *)o;
+#endif
+
+    if (channel_bindings != NULL) {
+        if (channel_bindings->initiator_address.value != NULL) {
+            PyMem_Free(channel_bindings->initiator_address.value);
+        }
+
+        if (channel_bindings->acceptor_address.value != NULL) {
+            PyMem_Free(channel_bindings->acceptor_address.value);
+        }
+
+        if (channel_bindings->application_data.value != NULL) {
+            PyMem_Free(channel_bindings->application_data.value);
+        }
+
+        free(channel_bindings);
+    }
+}
+
+static PyObject *channelBindings(PyObject *self, PyObject *args, PyObject* keywds) {
+    int initiator_addrtype = GSS_C_AF_UNSPEC;
+    int acceptor_addrtype = GSS_C_AF_UNSPEC;
+
+    const char *encoding = NULL;
+    char *initiator_address = NULL;
+    char *acceptor_address = NULL;
+    char *application_data = NULL;
+    int initiator_length = 0;
+    int acceptor_length = 0;
+    int application_length = 0;
+
+    PyObject *pychan_bindings = NULL;
+    struct gss_channel_bindings_struct *input_chan_bindings;
+    static char *kwlist[] = {"initiator_addrtype", "initiator_address", "acceptor_addrtype",
+        "acceptor_address", "application_data", NULL};
+
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "|iet#iet#et#", kwlist,
+            &initiator_addrtype, &encoding, &initiator_address, &initiator_length,
+            &acceptor_addrtype, &encoding, &acceptor_address, &acceptor_length,
+            &encoding, &application_data, &application_length)) {
+        return NULL;
+    }
+
+    input_chan_bindings = (struct gss_channel_bindings_struct *) malloc(sizeof(struct gss_channel_bindings_struct));
+    pychan_bindings = PyNew(input_chan_bindings, &destruct_channel_bindings);
+
+    input_chan_bindings->initiator_addrtype = initiator_addrtype;
+    input_chan_bindings->initiator_address.length = initiator_length;
+    input_chan_bindings->initiator_address.value = initiator_address;
+
+    input_chan_bindings->acceptor_addrtype = acceptor_addrtype;
+    input_chan_bindings->acceptor_address.length = acceptor_length;
+    input_chan_bindings->acceptor_address.value = acceptor_address;
+
+    input_chan_bindings->application_data.length = application_length;
+    input_chan_bindings->application_data.value = application_data;
+
+    return Py_BuildValue("N", pychan_bindings);
+}
+
+static PyObject *authGSSClientStep(PyObject *self, PyObject *args, PyObject* keywds) {
     gss_client_state *state;
     PyObject *pystate;
     char *challenge = NULL;
+    PyObject *pychan_bindings = NULL;
+    struct gss_channel_bindings_struct *channel_bindings;
+    static char *kwlist[] = {"state", "challenge", "channel_bindings", NULL};
     int result = 0;
 
-    if (!PyArg_ParseTuple(args, "Os", &pystate, &challenge)) {
+    if (! PyArg_ParseTupleAndKeywords(args, keywds, "Os|O", kwlist, &pystate, &challenge, &pychan_bindings)) {
         return NULL;
     }
 
@@ -183,7 +252,17 @@ static PyObject *authGSSClientStep(PyObject *self, PyObject *args) {
         return NULL;
     }
 
-    result = authenticate_gss_client_step(state, challenge);
+    if (pychan_bindings == NULL) {
+        channel_bindings = GSS_C_NO_CHANNEL_BINDINGS;
+    } else {
+        if (!PyCheck(pychan_bindings)) {
+            PyErr_SetString(PyExc_TypeError, "Expected a gss_channel_bindings_struct object");
+            return NULL;
+        }
+        channel_bindings = PyGet(pychan_bindings, struct gss_channel_bindings_struct);
+    }
+
+    result = authenticate_gss_client_step(state, challenge, channel_bindings);
     if (result == AUTH_GSS_ERROR) {
         return NULL;
     }
@@ -506,6 +585,137 @@ static PyObject *authGSSServerTargetName(PyObject *self, PyObject *args) {
     return Py_BuildValue("s", state->targetname);
 }
 
+#ifdef GSSAPI_EXT
+static PyObject* authGSSWinRMEncryptMessage(PyObject* self, PyObject* args)
+{
+    char *input = NULL;
+    char *header = NULL;
+    int header_len = 0;
+    char *enc_output = NULL;
+    int enc_output_len = 0;
+    PyObject *pystate = NULL;
+    gss_client_state *state = NULL;
+    int result = 0;
+    PyObject *pyresult = NULL;
+
+    // NB: use et so we get a copy of the string (since gss_wrap_iov mutates it), and so we're certain it's always
+    // a UTF8 byte string
+    if (! PyArg_ParseTuple(args, "Oet", &pystate, "UTF-8", &input)) {
+        pyresult = NULL;
+        goto end;
+    }
+
+    if (!PyCheck(pystate)) {
+        PyErr_SetString(PyExc_TypeError, "Expected a context object");
+        pyresult = NULL;
+        goto end;
+    }
+
+    state = PyGet(pystate, gss_client_state);
+    if (state == NULL) {
+        pyresult = NULL;
+        goto end;
+    }
+
+    result = encrypt_message(state, input, &header, &header_len, &enc_output, &enc_output_len);
+
+    if (result == AUTH_GSS_ERROR) {
+        pyresult = NULL;
+        goto end;
+    }
+
+#if PY_MAJOR_VERSION >= 3
+    pyresult = Py_BuildValue("y# y#", enc_output, enc_output_len, header, header_len);
+#else
+    pyresult = Py_BuildValue("s# s#", enc_output, enc_output_len, header, header_len);
+#endif
+
+end:
+    if (input) {
+        PyMem_Free(input);
+    }
+    if (header) {
+        free(header);
+    }
+    if (enc_output) {
+        free(enc_output);
+    }
+
+    return pyresult;
+}
+
+static PyObject* authGSSWinRMDecryptMessage(PyObject* self, PyObject* args)
+{
+    char *header = NULL;
+    int header_len = 0;
+    char *enc_data = NULL;
+    int enc_data_len = 0;
+    PyObject *pystate = NULL;
+    PyObject *pyheader = NULL;
+    PyObject *pyenc_data = NULL;
+    gss_client_state *state = NULL;
+    char *dec_output = NULL;
+    int dec_output_len = 0;
+    int result = 0;
+    PyObject *pyresult = 0;
+
+    // NB: since the sig/data strings are arbitrary binary and don't conform to
+    // a valid encoding, none of the normal string marshaling types will work. We'll
+    // have to extract the data later.
+    if (! PyArg_ParseTuple(args, "OOO", &pystate, &pyenc_data, &pyheader)) {
+        pyresult = NULL;
+        goto end;
+    }
+
+    if (!PyCheck(pystate)) {
+        PyErr_SetString(PyExc_TypeError, "Expected a context object");
+        pyresult = NULL;
+        goto end;
+    }
+
+    state = PyGet(pystate, gss_client_state);
+    if (state == NULL) {
+        pyresult = NULL;
+        goto end;
+    }
+
+    // request the length and copy the header and encrypted input data from the Python strings
+    header_len = (int) PyBytes_Size(pyheader);
+    header = malloc(header_len);
+    memcpy(header, PyBytes_AsString(pyheader), header_len);
+
+    enc_data_len = (int) PyBytes_Size(pyenc_data);
+    enc_data = malloc(enc_data_len);
+    memcpy(enc_data, PyBytes_AsString(pyenc_data), enc_data_len);
+
+    result = decrypt_message(state, header, header_len, enc_data, enc_data_len, &dec_output, &dec_output_len);
+
+    if (result == AUTH_GSS_ERROR) {
+        pyresult = NULL;
+        goto end;
+    }
+
+#if PY_MAJOR_VERSION >= 3
+    pyresult = Py_BuildValue("y#", dec_output, dec_output_len);
+#else
+    pyresult = Py_BuildValue("s#", dec_output, dec_output_len);
+#endif
+
+end:
+    if (header) {
+        free(header);
+    }
+    if (enc_data) {
+        free(enc_data);
+    }
+    if (dec_output) {
+        free(dec_output);
+    }
+
+    return pyresult;
+}
+#endif
+
 static PyMethodDef KerberosMethods[] = {
     {"checkPassword",  checkPassword, METH_VARARGS,
      "Check the supplied user/password against Kerberos KDC."},
@@ -517,7 +727,7 @@ static PyMethodDef KerberosMethods[] = {
      "Initialize client-side GSSAPI operations."},
     {"authGSSClientClean",  authGSSClientClean, METH_VARARGS,
      "Terminate client-side GSSAPI operations."},
-    {"authGSSClientStep",  authGSSClientStep, METH_VARARGS,
+    {"authGSSClientStep",  (PyCFunction)authGSSClientStep, METH_VARARGS | METH_KEYWORDS,
      "Do a client-side GSSAPI step."},
     {"authGSSClientResponse",  authGSSClientResponse, METH_VARARGS,
      "Get the response from the last client-side GSSAPI step."},
@@ -531,11 +741,17 @@ static PyMethodDef KerberosMethods[] = {
      "Do a GSSAPI wrap."},
     {"authGSSClientUnwrap",  authGSSClientUnwrap, METH_VARARGS,
      "Do a GSSAPI unwrap."},
+    {"channelBindings",  (PyCFunction)channelBindings, METH_VARARGS | METH_KEYWORDS,
+     "Build the Channel Bindings Structure based on the input."},
 #ifdef GSSAPI_EXT
     {"authGSSClientWrapIov",  authGSSClientWrapIov, METH_VARARGS,
      "Do a GSSAPI iov wrap."},
     {"authGSSClientUnwrapIov",  authGSSClientUnwrapIov, METH_VARARGS,
      "Do a GSSAPI iov unwrap."},
+    {"authGSSWinRMEncryptMessage",  authGSSWinRMEncryptMessage, METH_VARARGS,
+     "Encrypt a message"},
+    {"authGSSWinRMDecryptMessage",  authGSSWinRMDecryptMessage, METH_VARARGS,
+     "Decrypt a message"},
 #endif
     {"authGSSServerClean",  authGSSServerClean, METH_VARARGS,
      "Terminate server-side GSSAPI operations."},
@@ -620,6 +836,29 @@ void initkerberos(void)
     PyDict_SetItemString(d, "GSS_MECH_OID_KRB5", PyNew(&krb5_mech_oid, NULL));
     PyDict_SetItemString(d, "GSS_MECH_OID_SPNEGO", PyNew(&spnego_mech_oid, NULL));
 
+    PyDict_SetItemString(d, "GSS_C_AF_UNSPEC", PyInt_FromLong(GSS_C_AF_UNSPEC));
+    PyDict_SetItemString(d, "GSS_C_AF_LOCAL", PyInt_FromLong(GSS_C_AF_LOCAL));
+    PyDict_SetItemString(d, "GSS_C_AF_INET", PyInt_FromLong(GSS_C_AF_INET));
+    PyDict_SetItemString(d, "GSS_C_AF_IMPLINK", PyInt_FromLong(GSS_C_AF_IMPLINK));
+    PyDict_SetItemString(d, "GSS_C_AF_PUP", PyInt_FromLong(GSS_C_AF_PUP));
+    PyDict_SetItemString(d, "GSS_C_AF_CHAOS", PyInt_FromLong(GSS_C_AF_CHAOS));
+    PyDict_SetItemString(d, "GSS_C_AF_NS", PyInt_FromLong(GSS_C_AF_NS));
+    PyDict_SetItemString(d, "GSS_C_AF_NBS", PyInt_FromLong(GSS_C_AF_NBS));
+    PyDict_SetItemString(d, "GSS_C_AF_ECMA", PyInt_FromLong(GSS_C_AF_ECMA));
+    PyDict_SetItemString(d, "GSS_C_AF_DATAKIT", PyInt_FromLong(GSS_C_AF_DATAKIT));
+    PyDict_SetItemString(d, "GSS_C_AF_CCITT", PyInt_FromLong(GSS_C_AF_CCITT));
+    PyDict_SetItemString(d, "GSS_C_AF_SNA", PyInt_FromLong(GSS_C_AF_SNA));
+    PyDict_SetItemString(d, "GSS_C_AF_DECnet", PyInt_FromLong(GSS_C_AF_DECnet));
+    PyDict_SetItemString(d, "GSS_C_AF_DLI", PyInt_FromLong(GSS_C_AF_DLI));
+    PyDict_SetItemString(d, "GSS_C_AF_LAT", PyInt_FromLong(GSS_C_AF_LAT));
+    PyDict_SetItemString(d, "GSS_C_AF_HYLINK", PyInt_FromLong(GSS_C_AF_HYLINK));
+    PyDict_SetItemString(d, "GSS_C_AF_APPLETALK", PyInt_FromLong(GSS_C_AF_APPLETALK));
+    PyDict_SetItemString(d, "GSS_C_AF_BSC", PyInt_FromLong(GSS_C_AF_BSC));
+    PyDict_SetItemString(d, "GSS_C_AF_DSS", PyInt_FromLong(GSS_C_AF_DSS));
+    PyDict_SetItemString(d, "GSS_C_AF_OSI", PyInt_FromLong(GSS_C_AF_OSI));
+    PyDict_SetItemString(d, "GSS_C_AF_X25", PyInt_FromLong(GSS_C_AF_X25));
+    PyDict_SetItemString(d, "GSS_C_AF_NULLADDR", PyInt_FromLong(GSS_C_AF_NULLADDR));
+
 error:
     if (PyErr_Occurred())
         PyErr_SetString(PyExc_ImportError, "kerberos: init failed");
