@@ -205,7 +205,7 @@ int authenticate_gss_client_clean(gss_client_state *state)
     return ret;
 }
 
-int authenticate_gss_client_step(gss_client_state* state, const char* challenge)
+int authenticate_gss_client_step(gss_client_state* state, const char* challenge, struct gss_channel_bindings_struct* channel_bindings)
 {
     OM_uint32 maj_stat;
     OM_uint32 min_stat;
@@ -238,7 +238,7 @@ int authenticate_gss_client_step(gss_client_state* state, const char* challenge)
                                     state->mech_oid,
                                     (OM_uint32)state->gss_flags,
                                     0,
-                                    NULL, //GSS_C_NO_CHANNEL_BINDINGS,
+                                    channel_bindings,
                                     &input_token,
                                     NULL,
                                     &output_token,
@@ -840,3 +840,85 @@ static void set_gss_error(OM_uint32 err_maj, OM_uint32 err_min)
 
     PyErr_SetObject(GssException_class, Py_BuildValue("((s:i)(s:i))", buf_maj, err_maj, buf_min, err_min));
 }
+
+#ifdef GSSAPI_EXT
+int encrypt_message(gss_client_state *state, char *message_input, char **header, int *header_len, char **encrypted_data, int *encrypted_data_len)
+{
+    OM_uint32 maj_stat;
+    OM_uint32 min_stat;
+    int ret;
+    char *outloc = NULL;
+
+    gss_iov_buffer_desc iov[3];
+    iov[0].type = GSS_IOV_BUFFER_TYPE_HEADER | GSS_IOV_BUFFER_FLAG_ALLOCATE;
+    iov[1].type = GSS_IOV_BUFFER_TYPE_DATA;
+    // NB: this will modify the message buffer in place, ensure caller has copied the string if necessary
+    iov[1].buffer.value = message_input;
+    iov[1].buffer.length = strlen(message_input);
+    iov[2].type = GSS_IOV_BUFFER_TYPE_PADDING | GSS_IOV_BUFFER_FLAG_ALLOCATE;
+
+    maj_stat = gss_wrap_iov(&min_stat, state->context, 1, GSS_C_QOP_DEFAULT, NULL, iov, 3);
+
+    if (GSS_ERROR(maj_stat)) {
+        set_gss_error(maj_stat, min_stat);
+        ret = AUTH_GSS_ERROR;
+        goto end;
+    }
+
+    *header_len = iov[0].buffer.length;
+    *header = malloc(*header_len);
+    memcpy(*header, iov[0].buffer.value, *header_len);
+
+    // copy encrypted data, concatenate padding buffer if present
+    *encrypted_data_len = iov[1].buffer.length + iov[2].buffer.length;
+    *encrypted_data = malloc(*encrypted_data_len);
+    outloc = *encrypted_data;
+    outloc = mempcpy(outloc, iov[1].buffer.value, iov[1].buffer.length);
+    // NB: no-op if no padding is necessary (which seems to always be the case with aes256-cts-hmac-sha1-96)
+    mempcpy(outloc, iov[2].buffer.value, iov[2].buffer.length);
+
+    ret = 0;
+end:
+    maj_stat = gss_release_iov_buffer(&min_stat, iov, 3);
+
+    return ret;
+}
+
+int decrypt_message(gss_client_state *state, char *header, int header_len, char *data, int data_len, char **decrypted_output, int *decrypted_output_len)
+{
+    OM_uint32 maj_stat;
+    OM_uint32 min_stat;
+    int ret = 0;
+    int conf_state;
+    gss_qop_t qop_state;
+
+    // NB: ensure the caller has copied the python input buffer
+    // so we're not mutating somebody else's string
+
+    gss_iov_buffer_desc iov[3];
+    iov[0].type = GSS_IOV_BUFFER_TYPE_HEADER;
+    iov[0].buffer.value = header;
+    iov[0].buffer.length = header_len;
+
+    iov[1].type = GSS_IOV_BUFFER_TYPE_DATA;
+    iov[1].buffer.value = data;
+    iov[1].buffer.length = data_len;
+
+    maj_stat = gss_unwrap_iov(&min_stat, state->context, &conf_state, &qop_state, iov, 2);
+
+    if (GSS_ERROR(maj_stat)) {
+        set_gss_error(maj_stat, min_stat);
+        ret = AUTH_GSS_ERROR;
+        goto end;
+    }
+
+    *decrypted_output = malloc(iov[1].buffer.length);
+    *decrypted_output_len = iov[1].buffer.length;
+
+    memcpy(*decrypted_output, iov[1].buffer.value, iov[1].buffer.length);
+
+    ret = 0;
+end:
+    return ret;
+}
+#endif
\ No newline at end of file
